# Copyright 2010, 2013 Red Hat Inc.
# Author: Shreyank Gupta <sgupta@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import xmlrpclib
import cookielib
import urllib2
import httplib
import socket
import errno
import os


from classes import Bug, Component, User
from utils import extract, attachment_encode

BUGZILLA_URL = 'https://bugzilla.redhat.com/xmlrpc.cgi'
COOKIE_DIR = '/tmp/bzcookies/'


USER_AGENT = 'Python-urllib2/%s ' % urllib2.__version__

class BugzillaBase:
    """
    Base class
    Handles authentication, cookies and proxy.
    """

    def __init__(self,**kwargs):
        """
        Initialize a Bugzilla instance.
        
        Optional Arguments:
        -------------------
        url                     : The Bugzilla URL. 
                May or maynot end with /xmlrpc.cgi.
                If does not end with /xmlrpc.cgi, it will be assumed.
                If not provided, value of BUGZILLA_URL will be defaulted to.
        
        cookie_jar|cookiejar    : cookielib.CookieJar/MozillaCookieJar object.
        user|username|login     : Bugzilla login, usually an email id.
        password|passwd           : Password for bugzilla
        http_proxy|proxy        : String specifying the HTTP proxy of the
        bypass                  : boolean value, asks client to bypass 
                                    password auth and use cookies if present
                client's connection.
                Usually of the form server:port or http://server:port

        """
        # Initialize public attributes for unlogged unstance
        self.user_agent = USER_AGENT
        self.logged_in = False
        self.user_id = None

        self._init_private_data()
        # Extract provided values or default
        self._cookiejar = extract(kwargs, 'cookie_jar', 'cookiejar')
            
        self.url = extract(kwargs, 'url') or BUGZILLA_URL

        self.user = extract(kwargs, 'user', 'username', 'login') or ''
        self.password = extract(kwargs, 'password', 'passwd') or ''

        self.http_proxy = extract(kwargs, 'http_proxy', 'proxy') or ''
        self.bypass = extract(kwargs, 'bypass') or ''

        cookie_dir = extract(kwargs, 'cookie_dir') or COOKIE_DIR
        if not os.path.exists(cookie_dir):
            os.mkdir(cookie_dir)
        self.cookiefile = os.path.join(cookie_dir, '%s.cookie' % self.user)

        self.connect()

    def _init_private_data(self):
        """
        Initialize private variables used by this bugzilla instance.
        """
        self._cookiejar = None
        self._proxy = None
        self._transport = None

    #Methods for establishing bugzilla connection and logging in

    def initcookiefile(self, cookiefile=None):
        """
        Read the given (Mozilla-style) cookie file and fill in the
        cookiejar, allowing us to use saved credentials to access Bugzilla.
        If no file is given, self.cookiefile will be used.
        """
        if cookiefile: 
            self.cookiefile = cookiefile
        cj = cookielib.MozillaCookieJar(self.cookiefile)
        if os.path.exists(self.cookiefile):
            cj.load()
        else:
            # Create an empty cookiefile that's only readable by this user
            old_umask = os.umask(0077)
            cj.save(self.cookiefile)
            os.umask(old_umask)
        self._cookiejar = cj
        self._cookiejar.filename = self.cookiefile

    def connect(self, url='', http_proxy=''):
        """
        Connect to the bugzilla instance with the given url.
        
        If 'user' and 'password' are both set, we'll run login(). Otherwise
        you'll have to login() yourself before some methods will work.
        """
        if url:
            self.url = url
        if http_proxy:
            self.http_proxy = http_proxy
        if not self.http_proxy and os.environ.has_key('http_proxy'):
            self.http_proxy = os.environ['http_proxy']

        # Set up the transport
        self.initcookiefile() # sets _cookiejar
        if self.url.startswith('https'):
            self._transport = SafeCookieTransport()
        else:
            self._transport = CookieTransport() 
        self._transport.user_agent = self.user_agent
        self._transport.cookiejar = self._cookiejar
        
        # Set HTTP proxy if required
        if self.http_proxy:
            self._transport.set_proxy(self.http_proxy)

        # Assume ends with /xmlrpc.cgi
        if not self.url.endswith('xmlrpc.cgi'):
            self.url = urllib2.urlparse.urljoin(self.url, 'xmlrpc.cgi')

        # Set up the proxy, using the transport
        self._proxy = xmlrpclib.ServerProxy(self.url,self._transport,
            allow_none=True)
        # If cookies exist and password not provided, fake a login 
        if self._cookiejar._cookies and self.bypass:
            self._fake_login()
        elif self.user and self.password:
            self._logout()
            self.login()

    def disconnect(self):
        """
        Disconnect from the given bugzilla instance.
        """
        self._init_private_data() # clears all the connection state

    def _login(self,user,password):
        """
        Logs in using User.login
        """
        id = self._proxy.User.login({
            'login': user, 
            'password': password,
            })
        self.user_id = id['id']
        return self.user_id

    def _fake_login(self):
        """
        Fakes a login
        """
        self.logged_in = True
        self.password = ''

    def login(self,user=None,password=None):
        """
        Attempt to log in using the given username and password. Subsequent
        method calls will use this username and password. Returns False if 
        login fails, otherwise returns a numeric userid. It also sets the 
        logged_in attribute to True, if successful.

        If user is not set, the value of Bugzilla.user will be used. If *that*
        is not set, ValueError will be raised. 

        This method will be called implicitly at the end of connect() if user
        and password are both set. So under most circumstances you won't need
        to call this yourself.
        """
        if user:
            self.user = user
        if password:
            self.password = password

        if not self.user:
            raise ValueError, "missing username"
        if not self.password:
            raise ValueError, "missing password"
           
        try: 
            r = self._login(self.user,self.password)
            self.logged_in = True
            self.password = ''
        except xmlrpclib.Fault, f:
            r = False
        return r

    def _logout(self):
        """
        Clears cookies
        """
        self._cookiejar.clear()
        self._cookiejar.save()

    def logout(self):
        """
        Log out of bugzilla. Drops server connection and user info, and
        destroys authentication cookies.
        """
        self._logout()
        self.user = ''
        self.password = ''
        self.user_id  = None
        self.logged_in  = False

    
    # XMLRPC fetch methods

    def get_bugs(self, ids, comments, attachments):
        """
        Accepts a list of int/str bug ids 
        Fetches a group of bugs and returns a list of Bug objects
        """
        extra_fields = ['flags', 'description']
        if comments:
            extra_fields.append('comments')
        if attachments:
            extra_fields.append('attachments')
        ids = {'ids' : [str(id) for id in ids], 
            'extra_fields': extra_fields}

        out = self._proxy.Bug.get(ids)
        return [Bug(bug, self) for bug in out['bugs']]


    def get_bug_upstream(self, id, comments=False, attachments=False):
        """
        Used upstream Bug.get to fetch bug details. It's faily limited in
        funtionality, and makes extra calls to fetch comments and attachments.
        Also does not fetch keywords and many such attributes. 

        Use get_bug until Bug.get is upto the mark. 
        """
        bugs = self.get_bugs([id])
        bug = bugs[0]
        if comments:
            bug.get_comments()
        if attachments:
            bug.get_attachments()
        return bug

    def get_bug(self, id, comments=False, attachments=False):
        """
        Uses Bug.get to fetch bug with keywords, 
        attachments and comments
        """
        return self.get_bugs([id], comments, attachments)[0]

    def create(self, **kwargs):
        """
        Creates a new bug from the given parameters

        Compulsory parameters:
        ----------------------
        product                         (string)
        component                       (string)
        summary                         (string)
        version                         (string)
        
        Defaulted parameters:
        ---------------------
        description                     (string)
        op_sys                          (string)
        platform                        (string)
        priority                        (string)
        severity                        (string)
        
        Optional parameters:
        --------------------
        url                             (string)
        whiteboard                      (string)
        assigned_to                     (string)
        qa_contact                      (string)
        target_milestone                (string)
        keywords                        (comma separated string/array)

        """
        if kwargs.has_key('keywords'):
            if isinstance(kwargs['keywords'], list):
                kwargs['keywords'] = ', '.join(kwargs['keywords'])
        out = self._proxy.Bug.create(kwargs)
        kwargs['id'] = out['id']
        kwargs['status'] = 'NEW'
        return Bug(kwargs, self)

    def update_bugs(self, ids, **kwargs):
        """
        Update multiple bugs with given parameters

        Accepted parameters:
        --------------------
        product                         (string)
            (also update component, version and target_milestone)
        component                       (string)
            (assigned_to, qa_contact gets updated)
        version                         (string)
        target_milestone                (string)
        op_sys                          (string)
        platform                        (string)
        summary                         (string)
        priority                        (string)
        severity                        (string)
        url                             (string)
        whiteboard                      (string)
        comment                         (string)
                commentprivacy          (boolean) 
        assigned_to                     (string)
        qa_contact                      (string)
        dupe_id                         (integer)
        status                          (string)
                resolution              (string)
        fixed_in                        (string)
        [add|remove]_groups              (array)
        [add|delete]_dependson          (array)
        [add|delete]_blocked            (array)
        [add|remove|set]_keywords (array)
        [add|delete]_partner            (array)
        [add|delete]_verified           (array)
        [add|delete]_cc                 (array)
        [add|delete]_alias              (array)
        [devel|qa|internal]_whiteboard  (string)
        [cclist|reporter]_accessible    (boolean)

        eg: 
        bz.update_bugs(
            [123456, 234567],
            product='Red Hat Enterprise Linux 6', 
            component='gimp',
            version='6.0', 
            target_milestone='rc', 
            add_group=['redhat']
            ) 

        """
        return self._update(ids, kwargs)

    def _update(self, ids, kwargs):
        """
        Private method to update bugs
        """
        # Hack to allow BZ attribute name changes to remain unaffected.
        pairs = [
            ('status', 'bug_status'),
            ('summary', 'short_desc'),
            ('platform', 'rep_platform'),
            ('severity', 'bug_severity'),
            ('whiteboard', 'status_whiteboard'),
            ]
        for p1, p2 in pairs:
            if p1 in kwargs:
                kwargs[p2] = kwargs[p1]
            if p2 in kwargs:
                kwargs[p1] = kwargs[p2]

        # Get default assignee and qa contact on component change
        if 'component' in kwargs:
            kwargs['reset_assigned_to'] = True
            kwargs['reset_qa_contact'] = True
        kwargs['ids'] = ids
        
        # Fix comments
        if 'comment' in kwargs:
            kwargs['comment'] = {'body': kwargs['comment']}
        if 'commentprivacy' in kwargs:
            kwargs['comment']['is_private'] = kwargs['commentprivacy']
            kwargs.pop('commentprivacy')
        
        # Fix keywords
        keywords = {}
        for each in ['add_keywords', 'delete_keywords', 'set_keywords']:
            if each in kwargs:
                keywords[each.split('_')[0]] = kwargs[each]
                kwargs.pop(each)
        if keywords:
            kwargs['keywords'] = keywords

        # Fix groups
        groups = {}
        for each in ['add_groups', 'delete_groups']:
            if each in kwargs:
                groups[each.split('_')[0]] = kwargs[each]
                kwargs.pop(each)
        if groups:
            kwargs['groups'] = groups 
        out =  self._proxy.Bug.update(kwargs)
        return out['bugs']

    def search(self, **kwargs):
        """
        Search Bugzilla, with complete bug attributes

        Accepted parameters:
        --------------------
        id                                          (integer)
        alias                                       (string)
        assigned_to                                 (string)
        qa_contact                                  (string)
        reporter                                    (string)
        summary                                     (string)
        product                                     (string)
        component                                   (string)
        version                                     (string)
        target_milestone                            (string)
        op_sys                                      (string)
        platform                                    (string)
        priority                                    (string)
        severity                                    (string)
        url                                         (string)
        whiteboard                                  (string)
        status                                      (string)
        resolution                                  (string)
        fixed_in                                    (string)

        limit - limit to <limit> bugs               (integer)
            offset - starting position for <limit>  (integer)
                                            
        """
        kwargs['exclude_fields'] = ['internals']
        kwargs['extra_fields'] = ['description']
        out = self._proxy.Bug.search(kwargs)
        return [Bug(bug, self) for bug in out['bugs']]

    def component(self, product_name, component_name):
        """
        Fetch component given product and component name
        Returns component object
        """
        hash = self._proxy.Component.get({
            'names': {
                'product': product_name,
                'component': component_name, 
                }
            })
        return Component(hash['components'][0])

    def get_components(self, product_id):
        """
        Fetch components for a given product id.
        """
        return self._get_components_multiple([product_id]).values()[0][0]

    def _get_components_multiple(self, ids):
        """
        Fetch details about multiple products given product id
        """
        return self._proxy.Product.get({'ids': ids})

    def _get_comments(self, ids):
        """
        Fetches comments for multiple bug ids
        """
        #TODO implement new_since
        ids = [str(id) for id in ids]
        out = self._proxy.Bug.comments({
            'ids': ids
            })
        return out['bugs']

    def _get_attachments(self, ids):
        """
        Fetches attachment info for multiple bug ids
        """
        ids = [str(id) for id in ids]
        out = self._proxy.Bug.attachments({
            'ids': ids
            })
        return out['bugs']

    def _get_flags(self, ids):
        """
        Fetches flags for multiple bug ids
        """
        out = self._proxy.Flag.get({
            'ids': ids
            })
        return out['bugs']

    def update_flags(self, ids, hash, nomail=False):
        """
        Updates flags for given bug ids
        hash is a dictionary with flagname as key and state as value 
        e.g. {
            'flag1': '+',
            'flag2': '?',
            }
        """
        updates = [{'name': name, 'status': status} for name, status in
            hash.iteritems()]
        out = self._proxy.Flag.update({
            'ids': ids,
            'nomail': nomail,
            'updates': updates,
            })
        return True

    def get_users(self, logins):
        """
        Fetches user details for multiple users
        Accepts a array of login emails
        """
        out = self._proxy.User.get({
            'names': logins
        })
        return [User(user) for user in out['users']]

    def get_user(self, login):
        """
        Fetches details for single user
        """
        users = self.get_users([login])
        return users[0]

    def create_user(self, login, name=None, password=None):
        """
        Create a user
        User trying this should have 'editusers' privilege.
        """
        in_hash = {
            'email': login
            }
        if name:
            in_hash['full_name'] = name
        if password:
            in_hash['password'] = password
        out = self._proxy.User.create(in_hash)
        return out['id']

    def update_users(self, logins, **kwargs):
        """
        Update single/multiple user(s). 
        User calling this method should have 'editusers' privilege.

        Pass string for logins if you wanna update a single user otherwise
        pass array of logins.

        Parameters:
        -----------
        logins - Single or array for login names            (string|array)
        name                                                (string)
        password                                            (string)
        [add|remove]_group - list of group names            (array of strings)

        e.g. update_users(
            'abc@xyz.com', 
            name='New Name',
            password='newpasswd',
            add_group=['g1', 'g2'],
            remove_group=['g3', 'g4']
            )
        """
        #name is changed to real_name
        if 'name' in kwargs:
            kwargs['real_name'] = kwargs['name']
            del kwargs['name']

        out = self._proxy.User.update({
            'names': logins,
            'updates': kwargs
            })
        return out['users_updates']

    def add_attachment(self, id, file, description, **kwargs):
        """
        Attach a file the bug ID. Returns the ID of the attachment
        or returns None if something goes wrong.
        file should be a filename.
        description is the short description of this attachment.
        Optional keyword args are as follows:
            filename:  this will be used as the filename for the attachment.
                       REQUIRED if attachfile is a file-like object with no
                       'name' attribute, otherwise the filename or .name
                       attribute will be used.
            comment:   An optional comment about this attachment.
            isprivate: Set to True if the attachment should be marked private.
            ispatch:   Set to True if the attachment is a patch.
            contenttype: The mime-type of the attached file. Defaults to
                         application/octet-stream if not set. NOTE that text
                         files will *not* be viewable in bugzilla unless you 
                         remember to set this to text/plain. So remember that!
        """
        f = open(file)
        kwargs['ids'] = id
        kwargs['summary'] = description
        if 'filename' not in kwargs:
            kwargs['file_name'] = os.path.basename(f.name)
        else:
            kwargs['file_name'] = kwargs['filename']
        if 'isprivate' not in kwargs:
            kwargs['is_private'] = False
        else:
            kwargs['is_private'] = kwargs['isprivate']
        if 'contenttype' not in kwargs:
            kwargs['content_type'] = 'application/octet-stream'
        else:
            kwargs['content_type'] = kwargs['contenttype']
        if not kwargs['content_type'] == 'text/plain':
            kwargs['data'] = attachment_encode(f)
        else:
            kwargs['data'] = f.read()
        out = self._proxy.Bug.add_attachment(kwargs)
        return out


    def add_comment(self, bug_id, comment, private=False):
        """
        Add a comment to a given bug id
        pass private=True for private comment
        """
        hash = {
            'id': bug_id,
            'comment': comment,
            }
        if private:
            hash['private'] = True
        out = self._proxy.Bug.add_comment(hash)
        if out:
            return out['id']
        else:
            return False


    def _fetch_url(self, url):
        """
        Fetches a url in BZ using cookies
        """
        cookie_p = urllib2.HTTPCookieProcessor(self._cookiejar)
        opener = urllib2.build_opener(cookie_p)
        return opener.open(url)
        

class CookieResponse:
    """
    Fake HTTPResponse object that we can fill with headers we got elsewhere.
    We can then pass it to CookieJar.extract_cookies() to make it pull out the
    cookies from the set of headers we have.
    """
    
    def __init__(self,headers): 
        self.headers = headers
    
    def info(self): 
        return self.headers


class CookieTransport(xmlrpclib.Transport):
    """
    A subclass of xmlrpclib.Transport that supports cookies.
    """
    cookiejar = None
    scheme = 'http'

    # Cribbed from xmlrpclib.Transport.send_user_agent 
    def send_cookies(self, connection, cookie_request):
        if self.cookiejar is None:
            self.cookiejar = cookielib.CookieJar()
        elif self.cookiejar:
            self.cookiejar.add_cookie_header(cookie_request)
            # Pull the cookie headers out of the request object...
            cookielist=list()
            for h,v in cookie_request.header_items():
                if h.startswith('Cookie'):
                    cookielist.append([h,v])
            # ...and put them over the connection
            for h,v in cookielist:
                connection.putheader(h,v)

    # This is the same request() method from xmlrpclib.Transport,
    # with a couple additions noted below

    def mod_request(self, host, handler, request_body, verbose=0):
        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        # ADDED: construct the URL and Request object for proper cookie handling
        request_url = "%s://%s%s" % (self.scheme,host,handler)
        cookie_request  = urllib2.Request(request_url) 

        self.send_request(h,handler,request_body)
        self.send_host(h,host) 
        self.send_cookies(h,cookie_request) # ADDED. creates cookiejar if None.
        self.send_user_agent(h)
        self.send_content(h,request_body)

        if hasattr(h, 'getreply'):
            # python 2.6 
            errcode, errmsg, headers = h.getreply()

            # ADDED: parse headers and get cookies here
            cookie_response = CookieResponse(headers)
            # Okay, extract the cookies from the headers
            self.cookiejar.extract_cookies(cookie_response,cookie_request)
            # And write back any changes
            if hasattr(self.cookiejar,'save'):
                try:
                    self.cookiejar.save(self.cookiejar.filename)
                except e:
                    print "Couldn't write cookiefile %s: %s" % \
                            (self.cookiejar.filename,str(e))

            if errcode != 200:
                raise xmlrpclib.ProtocolError(
                    "%s://%s%s" % (self.scheme,host,handler),
                    errcode, errmsg,
                    headers
                    )

            self.verbose = verbose

            try:
                sock = h._conn.sock
            except AttributeError:
                sock = None

            f = h.getfile()
            retval = self._parse_response(f, sock)
            return retval
        else:
            # python 2.7
            try:
                response = h.getresponse(buffering=True)

                # ADDED: parse headers and get cookies here
                cookie_response = CookieResponse(response.msg)
                # Okay, extract the cookies from the headers
                self.cookiejar.extract_cookies(cookie_response,cookie_request)
                # And write back any changes
                if hasattr(self.cookiejar,'save'):
                    try:
                        self.cookiejar.save(self.cookiejar.filename)
                    except Exception, e:
                        print "Couldn't write cookiefile %s: %s" % \
                                (self.cookiejar.filename,str(e))

                if response.status == 200:
                    self.verbose = verbose
                    return self.parse_response(response)
            except xmlrpclib.Fault:
                raise
            except Exception:
                # All unexpected errors leave connection in
                # a strange state, so we clear it.
                self.close()
                raise

        #discard any response data and raise exception
        if (response.getheader("content-length", 0)):
            response.read()
        raise xmlrpclib.ProtocolError(
            host + handler,
            response.status, response.reason,
            response.msg,
            )

    def request(self, host, handler, request_body, verbose=0):
        #retry request once if cached connection has gone cold
        for i in (0, 1):
            try:
                return self.mod_request(host, handler, request_body, verbose)
            except socket.error, e:
                if i or e.errno not in (errno.ECONNRESET, errno.ECONNABORTED):
                    raise
            except httplib.BadStatusLine: #close after we sent request
                if i:
                    raise

    # To enable proxy
    def set_proxy(self, http_proxy):
        self.http_proxy = http_proxy
        self.make_connection = self._make_proxy_connection
        self.send_request = self._send_proxy_request
        self.send_host = self._send_proxy_host
    
    def _make_proxy_connection(self, host):
        self.realhost = host
        h = httplib.HTTP(self.http_proxy)
        return h
        
    def _send_proxy_request(self, connection, handler, request_body):
        url = '%s://%s%s' % (self.scheme, self.realhost, handler)
        connection.putrequest("POST", url)

    def _send_proxy_host(self, connection, host):
        connection.putheader('Host', self.realhost)


class SafeCookieTransport(xmlrpclib.SafeTransport, CookieTransport):
    """
    SafeTransport subclass that supports cookies.
    """
    scheme = 'https'
    request = CookieTransport.request


class BugzillaLoginException(Exception):
    """
    Bugzilla exception for login failure.
    """
    def __init__(self, msg=''):
        self.msg = msg

    def __str__(self):
        return  "Login failure. %s" % self.msg




#!/usr/bin/python 
 
import glob 
import os.path 

if os.path.exists('/etc/techilla/techilla.conf'):
    conffile = glob.glob(os.path.join('/etc/techilla/techilla.conf')) 
else:
    conffile = glob.glob(os.path.join('techilla.conf')) 
for f in conffile:
    execfile(f)



# Copyright 2012 Red Hat Inc.
# Author: Kushal Das <kdas@redhat.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  See
# http://www.gnu.org/copyleft/gpl.html for the full text of the
# license.
#
from pprint import pprint
import sys
import os
from utils import ksc_set
from utils import run
var = sys.version[:3]
if var in ['2.2', '2.3', '2.4']:
    set = ksc_set # pragma: no cover

keywords = set(['auto','_Bool','break','case','char','_Complex','const','continue','default',
                'do','double','else','enum','extern','float','for','goto','if','_Imaginary','inline',
                'int','long','register','restrict','return','short','signed','sizeof','static','struct',
                'switch','typedef','union','unsigned','void','volatile','while','','main'])

endpoints = set(['~','!','@','#',
                 '$','%','^','&','*',')',
                 '+','|','>','<','?','/',
                 '.',',','"',"'",':',
                 ';','-','[',']','{','}','='])


def parse_c(path, mock=False):
    """
    Parse the given C file and
    return the list of symbols
    """
    # If the file does not exists
    # then return an empty list.
    if not os.path.isfile(path):
        return []
    try:
        # We are using cpp to parse out the comments.
        data = run('cpp -fpreprocessed %s' % path)
    except:
        print "Error in opening %s" % path
        return []
    lines = data.split('\n')
    result = []
    for line in lines:
        line = line.strip()
        word = ""
        w_flag = False
        if line.startswith('#' ):
            continue

        startstr = False
        for i,c in enumerate(line):
            if c == '"':
                if not startstr:
                    startstr = True
                else:
                    if i -1 != -1:
                        if line[i-1] != '\\':
                            startstr = False
                continue
            if startstr:
                continue
            if c.isspace():
                w_flag = True
                continue
            elif c in endpoints:
                word = ""
                continue
            elif c != '(':
                if w_flag:
                    word = ""
                    w_flag = False
                    word += c
                    continue
                word += c
            elif c == '(':
                word = word.strip()
                if not word:
                    continue # pragma: no cover
                if word in keywords or word.startswith("("):
                    word = ""
                    continue
                result.append(word)
                word = ""
                w_flag = False


    #pprint(set(result))
    return set(result)



# Copyright 2012 Red Hat Inc.
# Author: Kushal Das <kdas@redhat.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  See
# http://www.gnu.org/copyleft/gpl.html for the full text of the
# license.
#
"""This is the utility module for ksc
This contains various utility functions."""
import sys
from bz_xmlrpc import BugzillaBase
VAR = sys.version[:3]
#Try to import subprocess for latest pythons
try: # pragma: no cover
    import subprocess
except ImportError: # pragma: no cover
    #We don't have subprocess # pragma: no cover
    #We will use os.popen # pragma: no cover
    pass 

import os
import base64
import getpass
import xmlrpclib

#Data file directory
PATH = "/usr/share/ksc"

#whitelist directory
WHPATH = '/lib/modules'

def ksc_set(seq):
    """
    Internal implementaion set for older
    python versions
    """
    data = []
    for val in seq:
        if val not in data:
            data.append(val)
    return data


def ksc_walk(top, topdown=True, onerror=None, followlinks=False):
    """
    Internal implementation of os.walk for older
    python versions
    """
    try:
        names = os.listdir(top)
    except OSError, err: # pragma: no cover
        print err
        return

    dirs, nondirs = [], []
    for name in names:
        if os.path.isdir(os.path.join(top, name)):
            dirs.append(name)
        else:
            nondirs.append(name)

    res = []
    res.append([top, dirs, nondirs])
    for name in dirs:
        path = os.path.join(top, name)
        if followlinks or not os.path.islink(path): #pragma: no cover
            for fpath in ksc_walk(path, topdown, onerror, followlinks):
                res.append(fpath)
    return res


class Myfile:
    """
    Internal file like object
    """

    def __init__(self):
        self.data = ""

    def write(self, info):
        """
        Writes data
        """
        self.data = self.data + info

    def read(self):
        """
        Reads the data
        """
        return self.data


if VAR in ['2.2', '2.3', '2.4']: #pragma: no cover
    set = ksc_set
    WALK = ksc_walk
else:
    WALK = os.walk


def read_list(arch, kabipath, verbose=False):
    """
    Reads a whitelist file and returns the symbols
    """
    fpath = os.path.join(WHPATH, kabipath, "kabi_whitelist_%s" % arch)
    if not os.path.isfile(fpath): #pragma: no cover
        fpath = "data/kabi_whitelist_%s" % arch
    result = []
    try:
        if verbose: #pragma: no cover
            print "Reading %s" % fpath
        fptr = open(fpath)
        for line in fptr.readlines():
            if line.startswith("["):
                continue
            result.append(line.strip("\n\t"))
        fptr.close()
    except IOError, err: #pragma: no cover
        print err
        print "whitelist missing"
    return result


def read_total_list(arch):
    """
    Reads total symbol list and returns the list
    """
    fpath = os.path.join(PATH, "data/%s.all-kernel-symbols.txt" % arch)
    if not os.path.isfile(fpath):#pragma: no cover
        fpath = "data/%s.all-kernel-symbols.txt" % arch
    result = []
    try:
        fptr = open(fpath)
        for line in fptr.readlines():
            if line.startswith("["):
                continue #pragma: no cover
            result.append(line.strip("\n\t"))
        fptr.close()
    except IOError, err: #pragma: no cover
        print err
        print "Missing all symbol list"
    return result


def run(command):
    """
    runs the given command
    """
    if VAR in ['2.6', '2.7']:
        ret = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        out, _ = ret.communicate()
    else: #pragma: no cover
        fptr = os.popen(command)
        out = fptr.read()
    return out


def get_cfiles(path):
    """ This will return a list of C files"""
    image_list = []

    for root, _, files in WALK(path):
        for name in files:
            if name.lower().endswith('.c'):
                image_list.append(os.path.join(root, name))

    return image_list


def encode_base64(path):
    """
    Encode the file and return the data
    """
    chunksize = 3072
    data = ""
    try:
        fptr = open(path)
        if VAR in ['2.2', '2.3', '2.4']: #pragma: no cover
            myfile = Myfile()
            base64.encode(fptr, myfile)
            data = myfile.read()
        else:
            chunk = fptr.read(chunksize)
            while chunk:
                data = data + base64.b64encode(chunk)
                chunk = fptr.read(chunksize)
        fptr.close()
    except: # pragma: no cover
        print "Error in base64 encoding of %s" % path
    return data


def getconfig(path = '/etc/ksc.conf', mock = False):
    """
    Returns the bugzilla config
    """
    result = {}
    result['partner'] = ''
    result['partnergroup'] = 'partner-group'
    
    if not os.path.isfile(path):
        path = '/etc/ksc.conf'
    try:
        fptr = open(path)
        lines = fptr.readlines()
        fptr.close()
        for line in lines:
            if line.startswith("user="):
                result["user"] = line[5:-1]
            elif line.startswith("partner="):
                result["partner"] = line[8:-1]
            elif line.startswith("server="):
                result["server"] = line[7:-1]
            elif line.startswith("partnergroup="):
                result["group"] = line[13:-1]
        if 'user' not in result:
            print "User name is missing in configuration."
            return False
        if 'server' not in result or not result['server'].endswith('xmlrpc.cgi'):
            print "Servername is not valid in configuration."
            return False
        if not mock: # pragma: no cover
            if not result['partner'] or result['partner'] == 'partner':
                result["partner"] = raw_input("Partner name: ")
            print 'Current Bugzilla user: %s' % result['user']
            result['password'] = getpass.getpass('Please enter password: ')
        else:
            result['password'] = 'mockpassword'
        if not result['user']:
            print "Error: missing values in configuration file." 
            print "Bug not submitted"
            sys.exit(1)
    except Exception, err:
        print "Error reading %s" % path
        sys.exit(1)
    return result


def createbug(data, arch, mock=False, path='/etc/ksc.conf', releasename='6.5', filename=None):
    """
    Opens a bug in the Bugzilla
    """

    bughash = {'product': 'Red Hat Enterprise Linux 6'}
    bughash["component"] = 'kabi-whitelists'
    bughash["summary"] = "kABI Symbol Usage"
    bughash["version"] = releasename
    bughash["platform"] = arch
    bughash["severity"] = "medium"
    bughash["priority"] = "medium"
    bughash["description"] = "Creating the bug to attach the symbol"+\
           "usage details."
    groups = ["redhat"]
    
    #We change the path if only it is mock
    if mock:
        path = './data/ksc.conf'

    try:
        conf = getconfig(path, mock=mock)
        if not conf:
            return
        if 'group' in conf:
            if conf['group'] != 'partner-group':
                groups.append(conf['group'])
        bughash["groups"] = groups
        server = xmlrpclib.Server(conf['server'])
        bughash["Bugzilla_login"] = conf["user"]
        bughash["Bugzilla_password"] = conf["password"]
        bughash["cf_partner"] = [conf["partner"], ]
    except Exception, err:
        print "Problem is parsing the configuration."
        print err
        return
    if not mock: #pragma: no cover
        print "Creating a new bug"
    bugid = 0
    try:
        bugid = server.Bug.create(bughash)
        if not mock: #pragma: no cover
            print "Bug URL %s/show_bug.cgi?id=%s" % \
                  (conf['server'][:-11], bugid["id"])
            print "Attaching the report"

        dhash = {"filename": "ksc-result.txt"}
        dhash["contenttype"] = "text/plain"
        dhash["description"] = "kABI symbol usage."
        dhash["data"] = data

        try:
            server.bugzilla.addAttachment(int(bugid["id"]), \
                                      dhash, conf["user"], conf["password"])
        except:  # Now try 4.4 series of bugzilla calls
            bz = BugzillaBase(url=conf['server'], user=conf["user"], password=conf["password"])
            bz.add_attachment(int(bugid["id"]), filename, description="kABI symbol usage.", contenttype="text/plain")

        if not mock: #pragma: no cover
            print "Attached successfully"
    except Exception, err: #pragma: no cover
        print "Error: %s" % err
        print "Bug not submitted"
        sys.exit(1)

    return bugid['id']



#!/usr/bin/env python
# Copyright 2012 Red Hat Inc.
# Author: Kushal Das <kdas@redhat.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  See
# http://www.gnu.org/copyleft/gpl.html for the full text of the
# license.
#
import os
import sys
import tempfile
from pprint import pprint
from optparse import OptionParser
from utils import run, read_list
from utils import read_total_list, get_cfiles
from utils import encode_base64
from utils import createbug
from utils import ksc_set
from keywords import endpoints, keywords, parse_c

KSCVERSION = "ksc - Version 0.9.5"

var = sys.version[:3]
if var in ['2.2', '2.3', '2.4']:# pragma: no cover
    set = ksc_set

class Ksc(object):
    def __init__ (self, mock = False):
        """
        Init call
        """
        self.all_symbols_used = []
        self.nonwhite_symbols_used = []
        self.white_symbols = []
        self.matchdata = None
        self.total = None
        self.tmplist = []
        self.verbose = False
        self.mock = mock
        self.releasedir = None
        self.arch = None
        self.releasename = '6.5'

    def clean(self):
        self.all_symbols_used = []
        self.nonwhite_symbols_used = []
        self.white_symbols = []
        self.matchdata = None
        self.total = None


    def main(self, mock_options=None):
        """
        Main function for the logic
        """
        filename = os.path.join(os.path.expanduser("~"),"ksc-result.txt")
        #default architecture
        self.arch = "x86_64"

        parser = OptionParser()
        parser.add_option("-c", "--config", dest="config",
                          help="path to configuration file", metavar="CONFIG")        
        parser.add_option("-d", "--directory", dest="directory",
                          help="path to the directory", metavar="DIRECTORY")
        parser.add_option("-i", "--internal", action="store_true", dest="internal",
                          help="to create text files to be used internally ", metavar="INTERNAL")
        parser.add_option("-k", "--ko", dest="ko",
                          help="path to the ko file", metavar="KO")
        parser.add_option("-n", "--name", dest="releasename",
                          help="Red Hat release to file the bug against", metavar="RELEASENAME")
        parser.add_option("-p", "--previous", dest="previous",
                          help="path to previous resultset to submit as bug")
        parser.add_option("-r", "--release", dest="release",
                          help="RHEL whitelist release to compare against", metavar="RELEASE")
        parser.add_option("-s", "--submit",
                  action="store_true", dest="submit", default=False,
                  help="Submit to Red Hat Bugzilla")
        parser.add_option("-v", "--version",
                  action="store_true", dest="VERSION", default=False,
                  help="Prints KSC version number")



        if not self.mock: # pragma: no cover
            (options, args) = parser.parse_args(sys.argv[1:])
        else: # pragma: no cover
            (options, args) = parser.parse_args(mock_options)

        if options.VERSION:
            print KSCVERSION
            sys.exit(0)

        #Create the ksc.conf config path
        if options.config:
            path = os.path.abspath(os.path.expanduser(options.config))
        else:
            path = os.path.join(os.path.expanduser("~"),"ksc.conf")

        if options.previous: #Submit the result of previous run
            filename = os.path.abspath(os.path.expanduser(options.previous))
            if os.path.basename(filename) != 'ksc-result.txt':
                print "Please point to the ksc-result.txt file in -p option."
                return
            self.submit(filename, path)
            return

        self.releasedir = 'kabi-current'
        choices = { 'rhel6.0': 'kabi-rhel60', 'rhel6.1': 'kabi-rhel61', \
                    'rhel6.2': 'kabi-rhel62', 'rhel6.3': 'kabi-rhel63', 'rhel6.4': 'kabi-rhel64'}
        if options.release in choices:
            self.releasedir = choices[options.release]
        
        if options.ko:
            self.find_arch(options.ko)

        if options.internal:
            self.create_internal(options)
            return

        if options.directory:
            self.find_files(options.directory)
        elif options.ko: # pragma: no cover
            val = self.read_data(self.arch, self.releasedir)
            #Return if there is any issue in reading whitelists
            if not val:
                return
            self.parse_ko(options.ko)
            self.print_result()
            self.save_result([(self.arch,self.white_symbols, self.nonwhite_symbols_used)])
        else: #pragma: no cover
            print "You need to provide a path to a sources directory or .ko file."
            return

        #Now save the result

        if not options.submit:
            return

        if options.releasename:
            self.releasename = options.releasename


        if not self.mock: #pragma: no cover
            self.get_justification(filename)
        self.submit(filename, path)

    def create_internal(self, options):
        """
        Creates the text files for internal usage only.
        """
        if options.directory:
            file_list = get_cfiles(options.directory)
            result = []

            for arch in ['i686', 'x86_64', 's390x', 'ppc64']:
                self.clean()
                self.arch = arch
                val = self.read_data(self.arch, self.releasedir)
                #Return if there is any issue in reading whitelists
                if not val:
                    return
                for f in file_list:
                    self.tmplist = []
                    allsyms = parse_c(f, self.mock)
                    map(self.find_if, allsyms)

                result.append((self.arch, self.nonwhite_symbols_used))
            #now save the result
            self.save_result_internal(result)
        elif options.ko:  # pragma: no cover
            val = self.read_data(self.arch, self.releasedir)
            #Return if there is any issue in reading whitelists
            if not val:
                return
            self.parse_ko(options.ko)
            self.save_result_internal([(self.arch, self.nonwhite_symbols_used)])

    def save_result_internal(self, data):
        "Saves the data in text format for internal usage"

        for datum in data:
            filename = os.path.join(os.path.expanduser("~"), "ksc-internal-%s.yml" % datum[0])
            fobj = open(filename, 'w')
            for name in datum[1]:
                fobj.write('%s:"Enter justification text here."\n' % name)
            fobj.close()
            print filename

    def submit(self, filename, path):
        """
        Submits the resultset into Red Hat bugzilla.

        :arg filename: Full path the ksc-result.txt file.
        :arg path: Path to the config file.
        """
        if not self.mock: #Ask for user permission
            print "By using ksc to upload your data to Red Hat,",\
                             " you consent to Red Hat's receipt,"
            print "use and analysis of this data."
            while True:
                ans = raw_input("y/N:")
                if ans == 'N':
                    return
                elif ans == 'y':
                    break
        data = encode_base64(filename)
        createbug(data, self.arch, mock=self.mock, path=path, releasename=self.releasename, filename = filename)

    def get_justification(self, filename):
        """
        Get the justification from User
        on non-whitelist symbols
        
        """
        bold = "\033[1m"
        reset = "\033[0;0m"
        
        print bold
        print "On the next screen, the result log will be opened to allow"
        print 'you to provide technical justification on why these symbols need to'
        print 'be included in the KABI whitelist.'
        print 'Please provide sufficient information in the log, marked with the'
        print 'line below:'

        print "\nENTER JUSTIFICATION TEXT HERE\n" + reset
        print bold + 'Press ENTER for next screen.' + reset
        raw_input()
        
        editor = os.getenv('EDITOR')
        if editor:
            os.system(editor + ' ' + filename)
        else:
            os.system('vi '+ filename)
        return True

    def save_result(self, data):
        """
        Save the result in a text file
        """
        filename = os.path.join(os.path.expanduser("~"),"ksc-result.txt")
        try:
            f = open(filename, "w")
            for i, datum in enumerate(data):
                if i == 0:
                    command = "[command: %s]\n" % " ".join(sys.argv)
                else:
                    command = ""
                self.write_result(f, datum[0], command, datum[1], \
                                    datum[2])
            f.close()
            if not self.mock:
                print "A copy of the report is saved in %s" % filename

        except Exception, e:
            print "Error in saving the result file at %s" % filename
            print e


    def print_result(self):
        """
        Print the result (score)
        """
        total_len = len(set(self.all_symbols_used))
        non_white = len(set(self.nonwhite_symbols_used))
        white_len = float (len(set(self.white_symbols)))

        if total_len == 0: # pragma: no cover
            print "No kernel symbol usage found."
            return

        score = (white_len / total_len) * 100

        if not self.mock:
            print "Checking against architecture %s" % self.arch
            print "Total symbol usage: %s\t"\
                  "Total Non white list symbol usage: %s"\
                  % (total_len, non_white)
            print "Score: %0.2f%%\n" % score



    def find_arch(self, path):
        """
        Finds the architecture of the file in given path
        """
        try:
            rset = {'IBM S/390': 's390x', '64-bit PowerPC or cisco 7500': 'ppc64', 
                    'x86-64': 'x86_64', 'Intel 80386': 'i686'}
            data = run("file %s" % path)
            part = data.split(":")[1].split(",")[1]
            arch = rset.get(part.strip(), 'x86_64')
            self.arch = arch
        except:
            print "Invalid architecture, supported architectures are i686,x86_64,ppc64,s390x"
            sys.exit(1)


    def write_result(self, f, arch, command="", whitelist=[], nonwhitelist=[]):
        """
        Save the result set in the given file
        """
        try:
            if command:
                f.write("%s" % command)
            f.write("[%s]\n" % arch)
            f.write("[WHITELISTUSAGE]\n")
            for name in set(whitelist):
                f.write(name + '\n')
            f.write("[NONWHITELISTUSAGE]\n")
            for name in set(nonwhitelist):
                f.write('#' * 10)
                f.write('\n(%s)\n\nENTER JUSTIFICATION TEXT HERE\n\n' % name)
            if set(nonwhitelist):
                f.write('#' * 10)
                f.write('\n')
        except Exception, err:
            print err



    def read_data(self, arch, releasedir):
        """
        Read both data files
        """
        self.matchdata = read_list(arch, releasedir, self.verbose);
        self.total = read_total_list(arch);
        if self.matchdata == []:
            return False #Means exit now
        return True #Do not exit

    def parse_ko(self, path):
        """
        parse a ko file
        """
        if not os.path.isfile(path):
            print "KO file can not be found"
            return

        out = run("nm -u %s" % path )
        for line in out.split("\n"):
            data = line.split("U ")
            if len(data)  == 2:
                self.find_if(data[1])
        #self.print_tmplist(path)

    def find_if(self, name):
        """
        Find if the symbol is in whitelist or not
        """
        data = name.split('_R')
        if len(data) > 1:
            name = data[1]
        if name in self.matchdata:
            self.white_symbols.append(name)
            self.all_symbols_used.append(name)
        elif name in self.total:
            self.all_symbols_used.append(name)
            self.nonwhite_symbols_used.append(name)
            self.tmplist.append(name)



    def find_files(self, path):
        file_list = get_cfiles(path)
        result = []

        for arch in ['i686', 'x86_64', 's390x', 'ppc64']:
            self.clean()
            self.arch = arch
            val = self.read_data(self.arch, self.releasedir)
            #Return if there is any issue in reading whitelists
            if not val:
                return

            for f in file_list:
                self.tmplist = []
                allsyms = parse_c(f, self.mock)
                map(self.find_if, allsyms)
                #self.print_tmplist(f)

            self.print_result()
            result.append((self.arch,self.white_symbols, self.nonwhite_symbols_used))
        #now save the result
        self.save_result(result)
        #set the architecture of the bug as no-arch
        self.arch = "noarch"



if __name__ == '__main__':
    k = Ksc()
    k.main()
    sys.exit(0)



# Copyright 2010, 2012, 2013 Red Hat Inc.
# Author: Shreyank Gupta <sgupta@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

from datetime import datetime
import os
from utils import extract, to_datetime, show_bug_url

class Bug:
    """
    The Bugzilla Bug Object
    -----------------------

    Attributes:
    -----------
    bz                  : Points to the BugzillaBase instance for the bug

    id                  : bug id
    summary             : Bug summary
    description         : Bug description
    assigned_to         : Assigned to Bugzilla login
    qa_contact          : QA Contact Bugzilla login
    reporter            : Reported By Bugzilla login
    product             : Product
    component           : Component
    creation_time       : Bug creation time, datetime
    last_change_time    : Last modified time, datetime
    dupe_of             : Duplicate of bug id
    priority            : Bug Priority
    severity            : Bug Severity
    partner             : Partners - array
    target_milestone    : Target milestone
    status              : Bug Status
    resolution          : Resulotion of closed bug
    whiteboard          : Bug status whiteboard
    version             : Version of platform
    platform            : Platform architecture
    keywords            : Array of keywords
    fixed_in            : Fixed in version

    comments            : Array of Comment objects, incuding bug description.
    attachments         : Array of Attachment objects
    groups              : Array of Group objects
    flags               : Array of Flag objects

    """

    def __init__(self, hash, bz):
        """
        Initialise a bug object

        A hash of bug attributes and a BugzillaBase object needs to be passed
        as parameters.
        """

        self.bz = bz

        self.id = 0
        self.summary = ''
        self.description = ''
        self.assigned_to = ''
        self.qa_contact = ''
        self.reporter = ''
        self.product = ''
        self.component = ''
        self.creation_time = ''
        self.last_change_time = ''
        self.dupe_of = 0
        self.priority = ''
        self.severity = ''
        self.partner = ''
        self.target_milestone = ''
        self.status = ''
        self.whiteboard = ''
        self.resolution = ''
        self.version = 0
        self.platform = ''
        self.keywords = []
        self.fixed_in = ''
        self.comments = []
        self.attachments = []
        self.groups = []
        self.flags = []
        
        # Extract available data 
        if hash:
            self._populate(hash)

    def _be(self, *keys):
        """
        Private conviniance wrapper around extract. 
        Hash defaults to self._hash
        """
        return extract(self._hash, *keys)
        
    def _populate(self, hash):
        """
        Accepts a bug hash populates bug attributes
        """
        # Hack for searched bugs
        if 'internals' in hash:
            hash = hash['internals']
            hash['status'] = hash['status']['value']

        self._hash = hash

        self.id = self._be('id', 'bug_id') or self.id
        self.summary = self._be('summary', 'short_desc') or self.summary
        self.description = self._be('description') or self.description
        self.assigned_to = self._be('assigned_to') or self.assigned_to
        self.qa_contact = self._be('qa_contact') or self.qa_contact
        self.reporter = self._be('reporter', 'creator') or self.reporter
        self.product = self._be('product') or self.product
        self.component = self._be('component') or self.component
        self.creation_time = (to_datetime(self._be('creation_time'))
            or self.creation_time)
        self.last_change_time = (to_datetime(self._be('last_change_time')) 
            or self.last_change_time)

        self.dupe_of = self._be('dupe_of', 'dupe_id') or self.dupe_of
        self.priority = self._be('priority')  or self.priority
        self.severity = self._be('severity','bug_severity') or self.severity
        self.partner = self._be('cf_partner') or self.partner
        self.target_milestone = (self._be('target_milestone') or
            self.target_milestone)
        self.status = self._be('status', 'bug_status') or self.status
        self.whiteboard = (self._be('status_whiteboard', 'whiteboard')
            or self.whiteboard)
        self.resolution = self._be('resolution') or self.resolution
        self.version = self._be('version') or self.version
        self.platform = self._be('platform', 'rep_platform') or self.platform
        self.keywords = self._be('keywords') or self.keywords
        if isinstance(self.keywords, str):
            self.keywords = self.keywords.split(', ')
        self.fixed_in = self._be('fixed_in') or self.fixed_in

        self._comments = self._be('longdescs', 'comments')
        if self._comments:
            self._hash2comments(self._comments)

        self._attachments = self._be('attachments')
        if self._attachments:
             self._hash2attachments(self._attachments)

        self._groups = self._be('groups')
        if self._groups:
            self.groups = [_Group(group) for group in self._groups]

        self._flags = self._be('flags')
        if self._flags:
            self.flags = [Flag(flag) for flag in self._flags]

        if self.id and self.bz:
            self.url = show_bug_url(self.bz.url) + str(self.id)

    def _hash2comments(self, hash):
        self.comments = [Comment(self, each) for each in hash]
        if not self.description:
            self.description = self.comments[0].text

    def _hash2attachments(self, hash):
        self.attachments = [Attachment(self, each) for each in hash]

    def get_comments(self):
        """
        Fetches comments from Bugzilla, stores and returns them.

        For retriving comments without fetching them try the comments
        attribute 
        """
        comment_list = \
            self.bz._get_comments([self.id])[str(self.id)]['comments']
        self._hash2comments(comment_list)
        return self.comments

    def get_attachments(self):
        """
        Fetches attachment info for the bug from the bugzilla, stores and 
        returns them.

        For retriving attachments without fetching them try the attachments
        attribute 
        """
        attachment_list = \
            self.bz._get_attachments([self.id])[str(self.id)]
        self._hash2attachments(attachment_list)
        return self.attachments

    def update(self, **kwargs):
        """
        Update bug with parameters from kwargs

        Accepted parameters:
        --------------------
        product                         (string)
            (also update component, version and target_milestone)
        component                       (string)
            (assigned_to, qa_contact gets updated)
        version                         (string)
        target_milestone                (string)
        op_sys                          (string)
        platform                        (string)
        summary                         (string)
        priority                        (string)
        severity                        (string)
        url                             (string)
        whiteboard                      (string)
        comment                         (string)
                commentprivacy          (boolean) 
        assigned_to                     (string)
        qa_contact                      (string)
        dupe_id                         (integer)
        status                          (string)
                resolution              (string)
        fixed_in                        (string)
        [add|delete]_group              (array)
        [add|delete]_dependson          (array)
        [add|delete]_blocked            (array)
        [add|delete|makeexact]_keywords (array)
        [add|delete]_partner            (array)
        [add|delete]_verified           (array)
        [add|delete]_cc                 (array)
        [add|delete]_alias              (array)
        [devel|qa|internal]_whiteboard  (string)
        [cclist|reporter]_accessible    (boolean)

        eg: 
        bug.update(
            product='Red Hat Enterprise Linux 6', 
            component='gimp',
            version='6.0', 
            target_milestone='rc', 
            add_group=['redhat']
            ) 

        """
        out = self.bz._update(self.id, kwargs)
        if out:
            self._populate(kwargs)
            return True
        else:
            return False

    def close(self, resolution, comment=None, dupe_id=None):
        """
        Convenience wrapper around update to close bugs
        Optional arg: comment
        """
        kwargs = {
            'status': 'CLOSED',
            'resolution': resolution,
        }

        if comment:
            kwargs['comment'] = comment
        if resolution == 'DUPLICATE' and dupe_id:
            kwargs['dupe_of'] = dupe_id
        return self.update(**kwargs)

    def add_comment(self, comment, private=False):
        """
        Add a comment to the bug
        pass private=True for private comment
        """
        return self.bz.add_comment(self.id, comment, private)

    def get_groups(self):
        """
        Returns Bugzilla group names the bug is on, as a list
        """
        if self.groups:
            return [group.name for group in self.groups if group.ison]

    def _fetch_flags(self):
        """
        Fetches Bugzilla Flags for the bug and saves in the 'flags' attribute
        """
        out = self.bz._get_flags([self.id])
        self._flags = out[str(self.id)]['bug']
        if self._flags:
            self.flags = [Flag(flag) for flag in self._flags]

    def get_flags(self, fetch=False):
        """
        Return Bugzilla flags for the bug in dict format
        pass fetch=True in order to fetch the flags from Bugzilla
        
        NOTE/TODO: Does not return needinfo flags
        if fetch:
            self._fetch_flags()
        """

        return dict([(flag.name, flag.status) for flag in self.flags])

    def update_flags(self, hash):
        """
        Updates flags for the bug
        hash is a dictionary with flagname as key and state as value 
        e.g. {
            'flag1': '+',
            'flag2': '?',
            }

        NOTE/TODO: Does not work with needinfo flags
        """
        return self.bz.update_flags(self.id, hash)

    def add_attachment(self, file, description, **kwargs):
        """
        Attach a file to the bug. Returns the ID of the attachment
        or returns None if something goes wrong.

        Compulsory arguements:
        ----------------------
        file            : should be a filename.
        description     : is the short description of this attachment.

        Optional arguements:
        --------------------
        comment         : A comment about this attachment.
        isprivate       : Set to True if the attachment should be marked private.
        ispatch         : Set to True if the attachment is a patch.
        contenttype     : The mime-type of the attached file. Defaults to
                          application/octet-stream if not set. 
                     NOTE: text files will *not* be viewable in bugzilla 
                     unless you remember to set this to text/plain.
        """
        return self.bz.add_attachment(self.id, file, description, **kwargs)

    def refresh(self):
        """
        refresh bug values
        """
        self = self.bz.get_bug(self.id)


class Comment:
    """
    Bugzilla comment object
    """
    def __init__(self, bug, hash):
        """
        Initialize comments
        """
        self._hash = hash
        self.id = extract(hash, 'id', 'comment_id')
        self.author = extract(hash, 'email', 'author')
        self.bug = bug
        self.is_private = bool(extract(hash, 'is_private',
            'isprivate'))
        self.text = extract(hash, 'text', 'body')
        self.time = to_datetime(extract(hash, 'time', 'bug_when'))


class Attachment:
    """
    Bugzilla attachment object
    """
    def __init__(self, bug, hash):
        """
        Initialize attachments
        """
        self._hash = hash
        self.id = extract(hash, 'id', 'attach_id')
        self.content_type = extract(hash, 'content_type',
            'mimetype')
        self.creation_time = to_datetime(extract(hash,
            'creation_time', 'creation_ts'))
        self.attacher = extract(hash, 'attacher', 'submitter_id')
        self.description = extract(hash, 'description')
        self.file_name = extract(hash, 'file_name', 'filename')
        self.bug = bug
        self.is_private = bool(extract(hash, 'is_private',
            'isprivate'))
        self.is_obsolete = bool(extract(hash, 'is_obsolete',
            'isobsolete'))
        self.is_patch = bool(extract(hash, 'is_patch', 'ispatch'))
        self.is_url = bool(extract(hash, 'is_url', 'isurl'))
        self.last_change_time = to_datetime(extract(hash, 
            'last_change_time', 'modification_time'))
        self.fetch_url = ''
        if self.id and self.bug:
            self.fetch_url = bug.bz.url.replace('xmlrpc.cgi',
                'attachment.cgi?id=%s' % self.id)

    def fetch(self, path=None):
        """
        Fetches the attachment
        returns a urllib2 object of no download path is specified
        otherwise it returns the full path of the downloaded file.
        """
        stream = self.bug.bz._fetch_url(self.fetch_url)
        if path:
            full_path = os.path.join(path, self.file_name)
            f = open(full_path, 'w')
            f.write(stream.read())
            f.close()
            return full_path
        else:
            return stream


class _Group:
    """
    Internal class to implement Bugzilla groups
    """
    
    def __init__(self, hash):
        """
        Initialize
        """
        self._hash = hash
        if isinstance(hash, str):
            # Hack for searched bug groups
            self.name = hash
            self.ison = True
        else:
            self.bit = extract(hash, 'bit', 'id')
            self.name = extract(hash, 'name')
            self.description = extract(hash, 'description')
            self.ingroup = bool(extract(hash, 'ingroup'))
            self.ison = bool(extract(hash, 'ison'))
            self.mandatory = bool(extract(hash, 'mandatory'))
            self.othercontrol = bool(extract(hash, 'othercontrol'))
            self.direct = bool(extract(hash, 'direct'))
            self.isbuggroup = bool(extract(hash, 'isbuggroup'))
            self.userregexp = extract(hash, 'userregexp')


class Component:
    """
    Bugzilla product component object
    """

    def __init__(self, hash):
        """
        Initialize
        """
        self._hash = hash
        self.id = hash['id']
        self.name = hash['name']
        self.product_id = hash['product_id']
        self.product_name = hash['product_name']
        self.default_assignee = hash['default_assignee']
        self.default_cc = hash['default_cc']
        self.default_qa_contact = hash['default_qa_contact']
        self.description = hash['description']
        self.flags = hash['flags']

class User:
    """
    Bugzilla user object
    """

    def __init__(self, hash):
        """
        Initialize
        """
        self._hash = hash
        self.id = hash['id']
        self.email = hash['email']
        self.name = hash['real_name']
        self.groups = [_Group(group) for group in hash['groups']]

    def get_groups(self):
        """
        Returns group strings as a list
        """
        if self.groups:
            return [group.name for group in self.groups if group.direct]


class Flag:
    """
    Bugzilla flag object
    """

    def __init__(self, hash):
        """
        Initialize BZ flag object
        """
        self._hash = hash
        self.id = hash['id']
        self.name = hash['name']
        self.status = hash['status']
        self.setter = hash['setter']
        self.requestee = self._be('requestee')
        self.type_id = hash['type_id']
        self.last_modified = hash['modification_date']

    def _be(self, *keys):
        """
        Private conviniance wrapper around extract. 
        Hash defaults to self._hash
        """
        return extract(self._hash, *keys)
 



import unittest
import os
from utils import *
from keywords import *
from ksc import Ksc
import xmlrpclib
from mock import Mock, patch


class SetTest(unittest.TestCase):
    """
    To test the set operation
    """
    def runTest(self):
        l = ksc_set([1,2,3,4,1,2,3,2])
        assert len(l) == 4


class WalkTest(unittest.TestCase):
    """
    To test ksc walk function
    """
    def runTest(self):
        x = ksc_walk('./data')
        x = x[0][2]
        assert 'ksc.conf' in x


class Base64Test(unittest.TestCase):
    """
    Test the base64 encoding
    """
    def runTest(self):
        data = encode_base64('/bin/ls')


class ReadListTest(unittest.TestCase):
    """
    Test reading whitelist
    """
    def runTest(self):
        data = read_list("i686","")
        assert len(data) != 0

class ReadTotalListTest(unittest.TestCase):
    """
    Test reading all symbol names
    """
    def runTest(self):
        data = read_total_list("i686")
        assert len(data) != 0


class MyfileTest(unittest.TestCase):
    """
    To test the internal file like object
    """
    def runTest(self):
        mf = Myfile()
        mf.write("hello")
        assert "hello" == mf.read()


class FindCTest(unittest.TestCase):
    """
    To test finding C source code files
    """
    def runTest(self):
        data = get_cfiles('./data')
        assert len(data) == 1

class KeywordsTest(unittest.TestCase):
    """
    To view the keywords
    """
    def runTest(self):
        import keywords
        assert 'break'  in keywords.keywords
        assert 'register'  in keywords.keywords

class ParseCTest(unittest.TestCase):
    """
    To parse the C code
    """
    def runTest(self):
        data = parse_c('data/test.c')
        assert set(['printf', 'add_disk','add_drv','call_rcu_bh']) == data

class ParseCfailTest(unittest.TestCase):
    """
    To test our own set function
    """
    def runTest(self):
        data = parse_c('data/does_not_exists.c', True)
        assert [] == data

class RunCommandTest(unittest.TestCase):
    """
    To test our own set function
    """
    def runTest(self):
        data = run('uname -a')
        self.assertTrue(data.startswith('Linux'))

class GetConfigTest(unittest.TestCase):
    """
    To test our own set function
    """
    def runTest(self):
        data = getconfig('./data/ksc.conf', True)
        assert 'user' in data
        assert 'partner' in data
        assert 'server' in data

class CreateBugTest(unittest.TestCase):
    """
    Code to test createbug function
    """
    @patch('xmlrpclib.Server')
    def runTest(self, mock_server):
        s = mock_server.return_value
        s.Bug.create.return_value = {'id':110}
        s.bugzilla.addAttachment.return_value = True
        bugid = createbug('abcd','i686', True) #This is mock
        assert bugid == 110

class KSCTest(unittest.TestCase):
    """
    Code to test the main workflow class
    """
    @patch('xmlrpclib.Server')
    def runTest(self, mock_server):
        s = mock_server.return_value
        s.Bug.create.return_value = {'id':110}
        s.bugzilla.addAttachment.return_value = True
        k = Ksc(mock=True)
        k.main(['-r', 'rhel6.2', '-s', '-d', './data/'])

class ParseKOTest(unittest.TestCase):
    """
    Code to test parse_ko
    """
    @patch('ksc.run')
    def runTest(self, mock_run):
        mock_run.return_value = 'U add_disk\nU add_drv\nU call_rcu_bh'
        k = Ksc(mock=True)
        k.read_data('i686','kabi-current')
        k.parse_ko('./ksc.py')
        assert len(k.all_symbols_used) == 2
        assert len(k.nonwhite_symbols_used) == 1
        assert len(k.white_symbols) == 1


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""ksc"""
from distutils.core import setup
from distutils.core import Command
from unittest import TextTestRunner, TestLoader
import os


class TestCommand(Command):
    user_options = [ ]

    def initialize_options(self):
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        '''
        Finds all the tests modules in tests/, and runs them.
        '''
        testfiles = ['tests']
        #for t in glob(pjoin(self._dir, 'tests', '*.py')):
        #    if not t.endswith('__init__.py'):
        #        testfiles.append('.'.join(
        #            ['tests', splitext(basename(t))[0]])
        #        )

        tests = TestLoader().loadTestsFromNames(testfiles)
        t = TextTestRunner(verbosity = 1)
        t.run(tests)

data = []
for x in os.listdir('data/'):
    if x in ['ksc.conf','test.c']:
        continue
    elif x.startswith('kabi_whitelist_'):
        continue
    else:
        data.append('data/%s' % x)

setup(name='ksc',
      version='0.9.5',
      description="ksc tool",
      long_description = "Kernel Module Source Checker tool",
      cmdclass = { 'test': TestCommand},
      platforms = ["Linux"],
      author="Kushal Das",
      author_email="kdas@redhat.com",
      url="http://redhat.com",
      license = "http://www.gnu.org/copyleft/gpl.html",
      data_files=[("/usr/bin",['ksc']),
          ('/etc', ['data/ksc.conf']),
          ('/usr/share/ksc',['ksc.py','keywords.py','utils.py']),
          ('/usr/share/ksc/data', data),
          ('/usr/share/ksc/bz_xmlrpc', ['bz_xmlrpc/__init__.py','bz_xmlrpc/base.py','bz_xmlrpc/classes.py', 'bz_xmlrpc/settings.py', 'bz_xmlrpc/utils.py'])],

      )



# Copyright 2010, 2012 Red Hat Inc.
# Author: Shreyank Gupta <sgupta@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import base64
from datetime import datetime
import time

if hasattr(datetime, 'strptime'):
    #python 2.6
    strptime = datetime.strptime
else:
    #python 2.4 equivalent
    strptime = lambda date_string, format: datetime(*(time.strptime(date_string, format)[0:6]))


def attachment_encode(fh):
    """
    Ripped from python-bugzilla

    Return the contents of the file-like object fh in a form
    appropriate for attaching to a bug in bugzilla. This is the default
    encoding method, base64.
    """
    # Read data in chunks so we don't end up with two copies of the file
    # in RAM.
    chunksize = 3072 # base64 encoding wants input in multiples of 3
    data = ''
    chunk = fh.read(chunksize)
    while chunk:
        # we could use chunk.encode('base64') but that throws a newline
        # at the end of every output chunk, which increases the size of
        # the output.
        data = data + base64.b64encode(chunk)
        chunk = fh.read(chunksize)
    return data

def to_datetime(xmlrpc_time):
    """
    Converts a xmlrpclib.DateTime object to a datetime.datetime object
    """
    if not xmlrpc_time:
        return None
    if isinstance(xmlrpc_time, str):
        try:
            return strptime(xmlrpc_time, "%Y-%m-%d %H:%M:%S")
        except:
            return strptime(xmlrpc_time, "%Y.%m.%d %H:%M")
    else:
        return strptime(xmlrpc_time.value, "%Y%m%dT%H:%M:%S")

def extract(hash, *keys):
    """
    Extracts dictionary value in order of key appearance
    """
    for key in keys:
        if hash.has_key(key):
            return hash[key]

def show_bug_url(xmlprc_url):
    """
    Returns BZ show_bug.cgi url from xmlrpc.cgi url
    """
    return xmlprc_url.replace('xmlrpc.cgi', 'show_bug.cgi?id=') 
